/*
 * IRCChannel.h
 * Object model of an IRC channel.
 *
 * Copyright (c) 1998 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1998.
 */

@class NSString, NSMutableDictionary;
@class IRCServer, IRCChannel, IRCEventProcessor, IRCUser;

/* channel modes */
#define IRCChannelAddMode			'+'
#define IRCChannelRemoveMode			'-'

#define IRCChannelPrivateMode			'p'
#define IRCChannelSecretMode			's'
#define IRCChannelModeratedMode			'm'
#define IRCChannelInviteOnlyMode		'i'
#define IRCChannelTopicOperatorOnlyMode		't'
#define IRCChannelClosedToExternalMessagesMode	'n'
#define IRCChannelKeyNeededMode			'k'
#define IRCChannelUserLimitedMode		'l'

#define IRCChannelOperatorMode			'o'
#define IRCChannelModeratorMode			'v'

#define IRCChannelBanMode			'b'

/* This notification is actually inherited from the IRCModel superclass */
IRC_EXPORT NSString *IRCChannelDidChangeNotification;

@interface IRCChannel : IRCModel <NSCopying>
{
    IRCServer *_server;
    NSString *_name;

    NSString *_topic;
    NSString *_topicSetByUserName;
    NSCalendarDate *_topicDate;

    NSMutableArray *_users;
    NSMutableArray *_operators;
    NSMutableArray *_moderators;

    struct channelFlags {
        BOOL isPrivate;
        BOOL isSecret;
        BOOL isModerated;
        BOOL isInviteOnly;
        BOOL isTopicOperatorOnly;
        BOOL isClosedToExternalMessages;
        BOOL isKeyNeeded;
        BOOL isUserLimited;
    } _channelFlags;

    unsigned _userLimit;
    NSString *_channelKey;
    NSMutableArray *_bans;

    // maintaining lists properly...
    BOOL _shouldFlushUsersOnNext353;
    BOOL _shouldFlushBansOnNext367;
}

+ (IRCChannel *)channelWithName:(NSString *)name server:(IRCServer *)server;
- (id)initWithName:(NSString *)name server:(IRCServer *)server;

/* retrieving channel information */
- (IRCServer *)server;

- (NSString *)topicSetByUserName;
- (NSCalendarDate *)topicDate;

- (unsigned)userLimit;
- (void)setUserLimit:(unsigned)userLimit;

- (NSString *)channelKey;
- (void)setChannelKey:(NSString *)channelKey;

- (NSString *)name;
- (NSString *)topic;

/* working with channel modes */
- (BOOL)isPrivate;
- (BOOL)isSecret;
- (BOOL)isModerated;
- (BOOL)isInviteOnly;
- (BOOL)isTopicOperatorOnly;
- (BOOL)isClosedToExternalMessages;
- (BOOL)isKeyNeeded;
- (BOOL)isUserLimited;

- (void)setPrivate:(BOOL)flag;
- (void)setSecret:(BOOL)flag;
- (void)setModerated:(BOOL)flag;
- (void)setInviteOnly:(BOOL)flag;
- (void)setTopicOperatorOnly:(BOOL)flag;
- (void)setClosedToExternalMessages:(BOOL)flag;
- (void)setKeyNeeded:(BOOL)flag;
- (void)setUserLimited:(BOOL)flag;

- (NSArray *)banList;
- (void)addBan:(NSString *)ban;
- (void)removeBan:(NSString *)ban;

/* user list management */
- (NSArray *)users;

- (void)addUser:(IRCUser *)user;
- (void)removeUser:(IRCUser *)user;
- (void)removeUserNamed:(NSString *)userName;

/* operators and moderators */
- (void)addOperator:(IRCUser *)user;
- (void)removeOperator:(IRCUser *)user;
- (BOOL)containsOperator:(IRCUser *)user;
- (BOOL)containsOperatorNamed:(NSString *)name;

- (void)addModerator:(IRCUser *)user;
- (void)removeModerator:(IRCUser *)user;
- (BOOL)containsModerator:(IRCUser *)user;
- (BOOL)containsModeratorNamed:(NSString *)name;

@end
