/* IRCChannel.m created by dwy on Wed 01-Apr-1998 */

#import "IRC.h"

DEFINE_NSSTRING(IRCChannelDidChangeNotification);

#define IRCChannelClassVersion		1

@implementation IRCChannel

+ (void)initialize
{
    if (self == [IRCChannel class])
        [self setVersion:IRCChannelClassVersion];
}

+ (IRCChannel *)channelWithName:(NSString *)name server:(IRCServer *)server
{
    if ([[server channelDictionary] objectForKey:name] == nil)
        [server addChannel:[[[self alloc] initWithName:name server:server] autorelease]];

    return [[server channelDictionary] objectForKey:name];
}

- (id)initWithName:(NSString *)name server:(IRCServer *)server
{
    if (self = [super init]) {
        _name = [name copy];
        _server = server;
        _users = [[NSMutableArray alloc] init];
        _operators = [[NSMutableArray alloc] init];
        _moderators = [[NSMutableArray alloc] init];
        _bans = [[NSMutableArray alloc] init];

        [[_server eventProcessor] addResponder:self];
    }

    return self;
}

- (void)dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];

    [_users release];
    [_operators release];
    [_moderators release];
    [_name release];
    [_bans release];
}

- (id)copyWithZone:(NSZone *)zone
{
    return [self retain];
}

- (NSComparisonResult)compare:(IRCChannel *)otherChannel
{
    return [_name isEqualToStringIgnoringCase:[otherChannel name]];
}

// some accessor methods
- (IRCServer *)server { return _server; }
- (NSString *)topicSetByUserName { return [[_topicSetByUserName copy] autorelease]; }
- (NSCalendarDate *)topicDate { return [[_topicDate copy] autorelease]; }
- (unsigned)userLimit { return _userLimit; }
- (NSString *)channelKey { return (_channelKey != nil) ? [[_channelKey copy] autorelease] : @""; }
- (NSString *)name { return [[_name copy] autorelease]; }
- (NSString *)topic { return (_topic ? [[_topic copy] autorelease] : @"No topic is set."); }
- (NSArray *)users { return _users; }

- (void)setUserLimit:(unsigned)userLimit { _userLimit = userLimit; [self registerInternalChangeWithNotificationCenter]; }
- (void)setChannelKey:(NSString *)channelKey
{
    [_channelKey release];
    _channelKey = [channelKey retain];
    [self registerInternalChangeWithNotificationCenter];
}

- (NSArray *)banList { return [[_bans copy] autorelease]; };

- (void)addBan:(NSString *)ban
{
    if (![_bans containsStringIgnoringCase:ban]) {
        [_bans addObject:ban];
        [self registerInternalChangeWithNotificationCenter];
    }
}

- (void)removeBan:(NSString *)ban
{
    if ([_bans containsStringIgnoringCase:ban]) {
        [_bans removeObject:ban];
        [self registerInternalChangeWithNotificationCenter];
    }
}

- (void)addUser:(IRCUser *)user
{
    [_users insertObject:user atIndex:0];
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(userDidChange:)
                                                 name:IRCUserDidChangeNotification
                                               object:user];
    [self registerInternalChangeWithNotificationCenter];
}

- (void)appendUser:(IRCUser *)user
{
    [_users addObject:user];
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(userDidChange:)
                                                 name:IRCUserDidChangeNotification
                                               object:user];
    [self registerInternalChangeWithNotificationCenter];
}

- (void)removeAllUsers
{
    [[NSNotificationCenter defaultCenter] removeObserver:self
                                                    name:IRCUserDidChangeNotification
                                                  object:nil];
    [_users removeAllObjects];
    _shouldFlushUsersOnNext353 = NO;
    [self registerInternalChangeWithNotificationCenter];
}

- (void)removeUser:(IRCUser *)user
{
    if (user != nil) {
        [self removeOperator:user];
        [self removeModerator:user];
        
        [[NSNotificationCenter defaultCenter] removeObserver:self
                                                        name:IRCUserDidChangeNotification
                                                      object:user];
        [_users removeObject:user];
        [self registerInternalChangeWithNotificationCenter];
    }
}

- (void)removeUserNamed:(NSString *)userName
{
    [self removeUser:[_users userNamed:userName]];
}

- (void)addOperator:(IRCUser *)user
{
    if (![_operators containsObject:user])
        [_operators addObject:user];

    [self registerInternalChangeWithNotificationCenter];
}

- (void)removeOperator:(IRCUser *)user
{
    if ([_operators containsObject:user])
        [_operators removeObject:user];

    [self registerInternalChangeWithNotificationCenter];
}    

- (BOOL)containsOperator:(IRCUser *)user { return [_operators containsObject:user]; }
- (BOOL)containsOperatorNamed:(NSString *)name { return [_operators containsUserNamed:name]; }

- (void)addModerator:(IRCUser *)user
{
    if (![_moderators containsObject:user])
        [_moderators addObject:user];

    [self registerInternalChangeWithNotificationCenter];
}

- (void)removeModerator:(IRCUser *)user
{
    if ([_moderators containsObject:user])
        [_moderators removeObject:user];

    [self registerInternalChangeWithNotificationCenter];
}    

- (BOOL)containsModerator:(IRCUser *)user { return [_moderators containsObject:user]; }
- (BOOL)containsModeratorNamed:(NSString *)name { return [_moderators containsUserNamed:name]; }

// flags methods
- (BOOL)isPrivate { return _channelFlags.isPrivate; }
- (BOOL)isSecret { return _channelFlags.isSecret; }
- (BOOL)isModerated { return _channelFlags.isModerated; }
- (BOOL)isInviteOnly { return _channelFlags.isInviteOnly; }
- (BOOL)isTopicOperatorOnly { return _channelFlags.isTopicOperatorOnly; }
- (BOOL)isClosedToExternalMessages { return _channelFlags.isClosedToExternalMessages; }
- (BOOL)isKeyNeeded { return _channelFlags.isKeyNeeded; }
- (BOOL)isUserLimited { return _channelFlags.isUserLimited; }

- (void)setPrivate:(BOOL)flag { _channelFlags.isPrivate = flag; [self registerInternalChangeWithNotificationCenter]; }
- (void)setSecret:(BOOL)flag { _channelFlags.isSecret = flag; [self registerInternalChangeWithNotificationCenter]; }
- (void)setModerated:(BOOL)flag { _channelFlags.isModerated = flag; [self registerInternalChangeWithNotificationCenter]; }
- (void)setInviteOnly:(BOOL)flag { _channelFlags.isInviteOnly = flag; [self registerInternalChangeWithNotificationCenter]; }
- (void)setTopicOperatorOnly:(BOOL)flag { _channelFlags.isTopicOperatorOnly = flag; [self registerInternalChangeWithNotificationCenter]; }
- (void)setClosedToExternalMessages:(BOOL)flag { _channelFlags.isClosedToExternalMessages = flag; [self registerInternalChangeWithNotificationCenter]; }
- (void)setKeyNeeded:(BOOL)flag { _channelFlags.isKeyNeeded = flag; [self registerInternalChangeWithNotificationCenter]; }
- (void)setUserLimited:(BOOL)flag { _channelFlags.isUserLimited = flag; [self registerInternalChangeWithNotificationCenter]; }

// main mode-handling routine
- (void)_processModeString:(NSString *)modeString arguments:(NSArray *)arguments
{
    char channelModeDirection = IRCChannelAddMode;
    unsigned argIndex = 0;
    const char *modePtr = [modeString cString];

    while (*modePtr) {
        switch (*modePtr) {
            case '+':	channelModeDirection = IRCChannelAddMode;	break;
            case '-':	channelModeDirection = IRCChannelRemoveMode;	break;

            case IRCChannelPrivateMode:	
                [self setPrivate:(channelModeDirection == IRCChannelAddMode)]; break;
            case IRCChannelSecretMode:
                [self setSecret:(channelModeDirection == IRCChannelAddMode)]; break;
            case IRCChannelModeratedMode:
                [self setModerated:(channelModeDirection == IRCChannelAddMode)]; break;
            case IRCChannelInviteOnlyMode:
                [self setInviteOnly:(channelModeDirection == IRCChannelAddMode)]; break;
            case IRCChannelTopicOperatorOnlyMode:
                [self setTopicOperatorOnly:(channelModeDirection == IRCChannelAddMode)]; break;
            case IRCChannelClosedToExternalMessagesMode:
                [self setClosedToExternalMessages:(channelModeDirection == IRCChannelAddMode)]; break;

            case IRCChannelKeyNeededMode:
                [self setKeyNeeded:(channelModeDirection == IRCChannelAddMode)];
                [self setChannelKey:((channelModeDirection == IRCChannelAddMode) ?
                                     [arguments objectAtIndex:argIndex++] :
                                     nil)];
                break;

            case IRCChannelUserLimitedMode:
                [self setUserLimited:(channelModeDirection == IRCChannelAddMode)];
                [self setUserLimit:((channelModeDirection == IRCChannelAddMode) ?
                                    [[arguments objectAtIndex:argIndex++] intValue] :
                                    0)];
                break;

            case IRCChannelOperatorMode:
                ((channelModeDirection == IRCChannelAddMode) ?
                 [self addOperator:[_users userNamed:[arguments objectAtIndex:argIndex++]]] :
                 [self removeOperator:[_users userNamed:[arguments objectAtIndex:argIndex++]]]);
                break;

            case IRCChannelModeratorMode:
                ((channelModeDirection == IRCChannelAddMode) ?
                 [self addModerator:[_users userNamed:[arguments objectAtIndex:argIndex++]]] :
                 [self removeModerator:[_users userNamed:[arguments objectAtIndex:argIndex++]]]);
                break;

            case IRCChannelBanMode:
                ((channelModeDirection == IRCChannelAddMode) ?
                 [self addBan:[arguments objectAtIndex:argIndex++]] :
                 [self removeBan:[arguments objectAtIndex:argIndex++]]);
                break;

            default:
                NSLog(@"unknown or unimplemented mode character %c", *modePtr);
                break;
        }

        modePtr++;
    }
}

- (void)userDidChange:(NSNotification *)notification
{
    [self registerInternalChangeWithNotificationCenter];
}

- (void)serverDid332:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:1] isEqualToStringIgnoringCase:_name]) {
        [_topic release];
        
        _topic = [[[event arguments] objectAtIndex:2] copy];
        [self registerInternalChangeWithNotificationCenter];
    }
}

- (void)serverDid333:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:1] isEqualToStringIgnoringCase:_name]) {
        [_topicSetByUserName release];
        [_topicDate release];
        
        _topicSetByUserName = [[[event arguments] objectAtIndex:2] copy];
        _topicDate = [[NSCalendarDate dateWithTimeIntervalSince1970:[[[event arguments] objectAtIndex:3] intValue]] retain];
        [self registerInternalChangeWithNotificationCenter];
    }
}

// initial channel-mode state; reset channel mode flags
- (void)serverDid324:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:1] isEqualToStringIgnoringCase:_name]) {
        _channelFlags.isPrivate = NO;
        _channelFlags.isSecret = NO;
        _channelFlags.isModerated = NO;
        _channelFlags.isInviteOnly = NO;
        _channelFlags.isTopicOperatorOnly = NO;
        _channelFlags.isClosedToExternalMessages = NO;
        _channelFlags.isKeyNeeded = NO;
        _channelFlags.isUserLimited = NO;
        _userLimit = 0;
        [_channelKey release];
        _channelKey = nil;

        [self _processModeString:[[event arguments] objectAtIndex:2]
                       arguments:[[event arguments] subarrayWithRange:NSMakeRange(3, [[event arguments] count]-3)]];
    }
}

- (void)serverDid353:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:2] isEqualToStringIgnoringCase:_name]) {
        NSEnumerator *userNameEnumerator = [[[[event arguments] objectAtIndex:3] componentsSeparatedByString:@" "] objectEnumerator];
        NSString *userName;

        if (_shouldFlushUsersOnNext353) {
            [self removeAllUsers];
            _shouldFlushUsersOnNext353 = NO;
        }

        while (userName = [userNameEnumerator nextObject]) {
            BOOL shouldOpUser = NO;
            BOOL shouldVoiceUser = NO;

            if ([userName length] > 0) {
                IRCUser *newUser = nil;

                if ([userName hasPrefix:IRCUserVoicePrefix]) {
                    shouldVoiceUser = YES;
                    userName = [userName substringFromIndex:1];
                }
                else if ([userName hasPrefix:IRCUserChannelOperatorPrefix]) {
                    shouldOpUser = YES;
                    userName = [userName substringFromIndex:1];
                }

                newUser = [IRCUser userWithNickname:userName server:server];
                [self appendUser:newUser];
                if (shouldOpUser)
                    [self addOperator:newUser];
                if (shouldVoiceUser)
                    [self addModerator:newUser];
            }
        }
        
        [self registerInternalChangeWithNotificationCenter];
    }
}

// End of /NAMES list.
- (void)serverDid366:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:1] isEqualToStringIgnoringCase:_name]) {
        _shouldFlushUsersOnNext353 = YES;
    }
}

// single ban item
- (void)serverDid367:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:1] isEqualToStringIgnoringCase:_name]) {
        if (_shouldFlushBansOnNext367) {
            [_bans removeAllObjects];
            _shouldFlushBansOnNext367 = NO;
        }

        [self addBan:[[event arguments] objectAtIndex:2]];
    }
}

// End of Channel Ban list.
- (void)serverDid368:(IRCServer *)server event:(IRCEvent *)event
{
   if ([[[event arguments] objectAtIndex:1] isEqualToStringIgnoringCase:_name]) {
       _shouldFlushBansOnNext367 = YES;
   }
}

- (void)serverDidJOIN:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:0] isEqualToStringIgnoringCase:_name]) {
        // if it's us joining, find out the channel modes and ban list
        if ([[[event prefix] nicknameComponent] isEqualToStringIgnoringCase:[[_server localUser] nickname]]) {
            [_server sendEvent:[IRCEvent modeEventWithTarget:_name
                                                       modes:[NSArray array]
                                                   arguments:[NSArray array]]];
            [_server sendEvent:[IRCEvent modeEventWithTarget:_name
                                                       modes:[NSArray arrayWithObject:@"b"]
                                                   arguments:[NSArray array]]];
        }
        else { // if it's someone else, add them to the user list.
            [self addUser:[IRCUser userWithNickname:[[event prefix] nicknameComponent]
                                             server:_server]];
        }
    }
}

// schedule these for next run-loop so that other responders may respond,
// e.g. channels, since the only way for a channel document to check if a quit
// applies too them is through this object!
- (void)serverDidKICK:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:0] isEqualToStringIgnoringCase:_name])
        [self performSelector:@selector(removeUserNamed:)
                   withObject:[[event arguments] objectAtIndex:1]
                   afterDelay:0.0];
}

- (void)serverDidPART:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:0] isEqualToStringIgnoringCase:_name])
        [self performSelector:@selector(removeUserNamed:)
                   withObject:[[event prefix] nicknameComponent]
                   afterDelay:0.0];
}

- (void)serverDidQUIT:(IRCServer *)server event:(IRCEvent *)event
{
    if ([_users containsUserNamed:[[event prefix] nicknameComponent]])
        [self performSelector:@selector(removeUserNamed:)
                   withObject:[[event prefix] nicknameComponent]
                   afterDelay:0.0];
}

- (void)serverDidTOPIC:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:0] isEqualToStringIgnoringCase:_name]) {
        [_topic release];
        _topic = [[[event arguments] objectAtIndex:1] copy];
        _topicSetByUserName = [[[event prefix] nicknameComponent] retain];
        _topicDate = [[NSDate date] retain];
        
        [self registerInternalChangeWithNotificationCenter];
    }
}

- (void)serverDidMODE:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:0] isEqualToStringIgnoringCase:_name])
        [self _processModeString:[[event arguments] objectAtIndex:1]
                       arguments:[[event arguments] subarrayWithRange:NSMakeRange(2, [[event arguments] count]-2)]];
}

- (void)serverWillTerminate:(NSNotification *)notification
{
    [[_server eventProcessor] removeResponder:self];
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] name: %@ userLimit: %u key: \"%@\" topic: \"%@\" setBy: date: %@ %@ users: %@>",
        [[self class] description], self,
        _name, _userLimit, _channelKey,
        _topic, _topicSetByUserName, [_topicDate description], [_users description]];
}

@end
