/*
 * IRCEvent.m
 * Model communications to an IRC server.
 *
 * Copyright (c) 1998 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1998.
 */

#import "IRC.h"

NSString *IRCCTCPDelimiterCharacter = @"\001";

@implementation IRCEvent

+ eventWithMessage:(NSString *)aString
{
    return [[[self alloc] initWithMessage:aString] autorelease];
}

- (id)initWithPrefix:(NSString *)prefix command:(NSString *)command arguments:(NSArray *)arguments
{
    if (self = [super init]) {
        _prefix = [prefix copy];
        _command = [command copy];
        _arguments = [arguments copy];
    }

    return self;
}

- (id)initWithMessage:(NSString *)message
{
    NSString *prefix = nil, *command = nil;
    NSMutableArray *arguments = [NSMutableArray array];
    NSScanner *scanner = [NSScanner scannerWithString:message];

    // parse the prefix
    if ([message hasPrefix:@":"]) {
        [scanner scanString:@":" intoString:NULL];
        [scanner scanUpToString:@" " intoString:&prefix];
    }

    // parse the command
    [scanner scanUpToString:@" " intoString:&command];
    
    while (![scanner isAtEnd]) {
        NSString *temp2;

        if ([scanner scanString:@":" intoString:NULL]) {
            [scanner scanUpToString:NULL intoString:&temp2];
            [arguments addObject:temp2];
        }
        else if ([scanner scanUpToString:@" " intoString:&temp2]) {
            [arguments addObject:temp2];
            [scanner scanString:@" " intoString:NULL];
        }
        else
            [NSException raise:NSGenericException
                        format:@"unparsable event! %@", message];
    }
    
    return [self initWithPrefix:prefix command:command arguments:arguments];
}

- (NSString *)prefix
{
    return [[_prefix copy] autorelease];
}

- (NSString *)command;
{
    return [[_command copy] autorelease];
}

- (NSArray *)arguments;
{
    return [[_arguments copy] autorelease];
}

- (SEL)selector
{
    return NSSelectorFromString([NSString stringWithFormat:@"serverDid%@:event:", _command]);
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] prefix: \"%@\" command: \"%@\" arguments: %@>",
        [[self class] description], self,
        _prefix, _command, [_arguments description]];
}

- (NSData *)encodedData
{
    return [[self string] dataUsingEncoding:NSASCIIStringEncoding];
}

- (NSString *)string
{
    NSMutableString *string = [NSMutableString new];
    NSEnumerator *argumentEnumerator = [[self arguments] objectEnumerator];
    NSString *argument;

    // prefices are optional
    if (_prefix != nil)
        [string appendString:[NSString stringWithFormat:@"%@ ", _prefix]];

    // commands are not
    if (_command != nil)
        [string appendString:_command];
    else
        [NSException raise:NSGenericException
                    format:@"IRCEvent may not have nil _command"];

    while (argument = [argumentEnumerator nextObject]) {
        if ([argument rangeOfString:@" "].length == 0) {
            [string appendString:@" "];
            [string appendString:argument];
        }
        else {
            NSMutableArray *arguments = [NSMutableArray arrayWithObject:argument];

            [arguments addObjectsFromArray:[argumentEnumerator remainingObjects]];
            [string appendString:[NSString stringWithFormat:@" :%@",
                [arguments componentsJoinedByString:@" "]]];
            
            break;
        }
    }

    [string appendString:@"\n"];

    return [[string copy] autorelease];
}

- (void)dealloc
{
    [_prefix release];
    [_command release];
    [_arguments release];

    [super dealloc];
}

- (id)copyWithZone:(NSZone *)zone
{
    return [self retain];
}

- (id)initWithCoder:(NSCoder *)coder
{
    if (self = [super init]) {
        _prefix = [[coder decodeObject] retain];
        _command = [[coder decodeObject] retain];
        _arguments = [[coder decodeObject] retain];
    }

    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [coder encodeObject:_prefix];
    [coder encodeObject:_command];
    [coder encodeObject:_arguments];
}

@end
