/*
 * IRCEventProcessor.m
 * Given an IRC server, perform relevant I/O and manage the event responder chain.
 *
 * Copyright (c) 1998 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1998.
 */

#import "IRC.h"

@implementation IRCEventProcessor

- (id)initWithServer:(IRCServer *)server
{    
    if (self = [super init]) {
        _server = server;
        _socket = [server socket];

        _responders = [[NSMutableArray alloc] init];

        _dataPort = [[NSPort alloc] init];
        [_dataPort setDelegate:self];
        [[NSRunLoop currentRunLoop] addPort:_dataPort forMode:NSDefaultRunLoopMode];

       [NSThread detachNewThreadSelector:@selector(readEventInBackground:)
                                 toTarget:self
                               withObject:_dataPort];
    }

    return self;
}

- (void)dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];

    [[NSRunLoop currentRunLoop] removePort:_dataPort forMode:NSDefaultRunLoopMode];
    [_dataPort release];

    [_responders release];

    [super dealloc];
}

- (void)addResponder:(id)responder
{
    if (![_responders containsObject:responder])
        [_responders addObject:responder];
}

- (void)removeResponder:(id)responder
{
    if ([_responders containsObject:responder])
        [_responders removeObject:responder];
}

- (NSArray *)responders
{
    return _responders;
}

- (void)dispatchEvent:(IRCEvent *)event responder:(id)responder
{
    SEL selector = [event selector];
    NSMethodSignature *signature = nil;

    if ([responder respondsToSelector:@selector(server:willProcessEvent:)])
        if ([responder server:_server willProcessEvent:event])
            return;

    signature = [responder methodSignatureForSelector:selector];
    if (signature != nil) {
        NSInvocation *invocation = [NSInvocation invocationWithMethodSignature:signature];

        [invocation setSelector:selector];
        [invocation setArgument:&_server atIndex:2];
        [invocation setArgument:&event atIndex:3];

        [invocation invokeWithTarget:responder];
    }
}

- (void)dispatchEvent:(IRCEvent *)event
{
    NSMutableArray *localResponders = [NSMutableArray array];
    NSEnumerator *responderEnumerator = [_responders objectEnumerator];
    id responder;

    while (responder = [responderEnumerator nextObject])
        [localResponders addObject:responder];

    responderEnumerator = [localResponders objectEnumerator];
    while (responder = [responderEnumerator nextObject])
        [self dispatchEvent:event responder:responder];
}

- (void)readEventInBackground:(NSPort *)dataPort
{
    NSAutoreleasePool *threadPool = [[NSAutoreleasePool alloc] init];
    NSString *line;
    
    NS_DURING {
        while (line = [_socket readLine]) {
            NSAutoreleasePool *subPool = [[NSAutoreleasePool alloc] init];
            IRCEvent *event;

            // strip \n, possible \r
            line = [line substringToIndex:[line length]-1];
            if ([line hasSuffix:@"\015"])
                line = [line substringToIndex:[line length]-1];
            
            event = [IRCEvent eventWithMessage:line];

            [dataPort sendData:[NSArchiver archivedDataWithRootObject:event]];
            [subPool release];
        }
    }
    NS_HANDLER {
        _threadException = [localException retain];
        [dataPort sendData:[NSData data]];
    }
    NS_ENDHANDLER;

    [threadPool release];
    [NSThread exit];
}

- (void)handlePortMessage:(NSPortMessage *)message
{
    NSData *data = [[message components] objectAtIndex:0];

    if ([data length] > 0)
        [self dispatchEvent:[NSUnarchiver unarchiveObjectWithData:data]];
    else
        [_server disconnect];
}

@end
