/*
 * IRCServer.m
 * Object model of an IRC server.
 *
 * Copyright (c) 1998 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1998.
 */

#import "IRC.h"

DEFINE_NSSTRING(IRCServerDidChangeNotification);
DEFINE_NSSTRING(IRCServerChannelListDidChangeNotification);
DEFINE_NSSTRING(IRCServerUserListDidChangeNotification);
DEFINE_NSSTRING(IRCServerDidConnectNotification);
DEFINE_NSSTRING(IRCServerDidRegisterNotification);
DEFINE_NSSTRING(IRCServerWillTerminateNotification);

id IRCApp;

void IRCSetApplicationObject(id anObject)
{
    IRCApp = [anObject retain];
}

@implementation IRCServer

+ (IRCServer *)serverWithHostName:(NSString *)hostName port:(int)port nickname:(NSString *)nickname userDescription:(NSString *)userDescription
{
    return [[[self alloc] initWithHostName:hostName
                                      port:port
                                  nickname:nickname
                           userDescription:userDescription] autorelease];
}

- (id)initWithHostName:(NSString *)hostName port:(int)port nickname:(NSString *)nickname userDescription:(NSString *)userDescription
{
    if (self = [super init]) {
        _description = @"No Description";
        _serverHost = [hostName retain];
        _serverPort = port;
        _channelDictionary = [[NSMutableDictionary alloc] init];
        _users = [[NSMutableArray alloc] init];

        _socket = [[TCSocketConnector defaultConnector] socketConnectedToHostName:_serverHost
                                                                             port:_serverPort];
        [_socket retain];

        _eventProcessor = [[IRCEventProcessor alloc] initWithServer:self];
        [_eventProcessor addResponder:self];

        [self sendEvent:[IRCEvent registrationEventWithUser:NSUserName()
                                                   userDescription:userDescription]];
        [self sendEvent:[IRCEvent nicknameChangeEventWithNickname:nickname]];

        [[NSNotificationCenter defaultCenter] postNotificationName:IRCServerDidConnectNotification
                                                            object:self];

        // create the local user object and add it to the user table cache.
        _localUser = [[IRCUser alloc] initWithNickname:nickname server:self];
        [self addUser:_localUser];
        [_localUser setUserDescription:userDescription];
    }

    return self;
}

- (TCSocket *)socket
{
    return _socket;
}

// this primitive should be rewritten to write asynchronously
- (void)sendData:(NSData *)data
{
    [_socket writeData:data];
}

- (void)sendEvent:(IRCEvent *)event
{
    [self sendData:[event encodedData]];
}

- (IRCEventProcessor *)eventProcessor
{
    return _eventProcessor;
}

- (BOOL)isConnected
{
    return (_socket != nil);
}

- (BOOL)isEqual:(id)anObject
{
    if ([anObject isMemberOfClass:[IRCServer class]])
        if ([[(IRCServer *)anObject hostName] isEqualToStringIgnoringCase:_serverHost])
            return YES;

    return NO;
}

- (void)setDescription:(NSString *)aDesc
{
    [_description release];
    _description = [aDesc retain];
    
    [self registerInternalChangeWithNotificationCenter];
}

- (IRCUser *)localUser
{
    return _localUser;
}

- (NSString *)hostName
{
    return [[_serverHost copy] autorelease];
}

- (int)port
{
    return _serverPort;
}

- (NSDictionary *)channelDictionary
{
    return _channelDictionary;
}

- (NSArray *)users
{
    return _users;
}

- (void)disconnect
{
    [[NSNotificationCenter defaultCenter]
	postNotificationName:IRCServerWillTerminateNotification
                      object:self];

    [_eventProcessor removeResponder:self];
    [_socket closeStream];
}

- (void)addChannel:(IRCChannel *)channel
{
    [_channelDictionary setObject:channel forKey:[channel name]];

    [[NSNotificationCenter defaultCenter]
        postNotificationName:IRCServerChannelListDidChangeNotification
                      object:self];
}

- (void)removeChannel:(IRCChannel *)channel;
{
    [self removeChannelNamed:[channel name]];
}

- (void)removeChannelNamed:(NSString *)channelName;
{
    [_channelDictionary removeObjectForKey:channelName];

    [[NSNotificationCenter defaultCenter]
        postNotificationName:IRCServerChannelListDidChangeNotification
                      object:self];
}

- (void)addUser:(IRCUser *)user
{
    [_users addObject:user];

    [[NSNotificationCenter defaultCenter]
        postNotificationName:IRCServerUserListDidChangeNotification
                      object:self];
}

- (void)removeUser:(IRCUser *)user
{
    [_users removeObject:user];
}

- (void)removeUserNamed:(NSString *)name
{
    [self removeUser:[_users userNamed:name]];

    [[NSNotificationCenter defaultCenter]
        postNotificationName:IRCServerUserListDidChangeNotification
                      object:self];
}

- (void)dealloc
{
    [_eventProcessor release];
    [_socket release];

    [_description release];
    [_serverHost release];
    [_channelDictionary release];
    [_users release];
    
    [_eventProcessor release];
    [_localUser release];

    [super dealloc];
}

// Our CTCP handler. This works by parsing out the CTCP and re-propagating the event through
// the responder chain. Isn't that elite? :)
- (void)serverDidPRIVMSG:(IRCServer *)server event:(IRCEvent *)event
{
    NSString *message = [[event arguments] objectAtIndex:1];
    if ([message hasPrefix:IRCCTCPDelimiterCharacter] && [message hasSuffix:IRCCTCPDelimiterCharacter]) {
        NSMutableArray *array = [NSMutableArray array];
        NSScanner *scanner = [NSScanner scannerWithString:message];
        NSString *ctcp = nil, *ctcpCommand = nil, *ctcpParameters = nil;
        IRCEvent *ctcpEvent;

        [scanner scanString:IRCCTCPDelimiterCharacter intoString:NULL];
        [scanner scanUpToString:IRCCTCPDelimiterCharacter intoString:&ctcp];
        [scanner scanString:IRCCTCPDelimiterCharacter intoString:NULL];

        scanner = [NSScanner scannerWithString:ctcp];
        [scanner scanUpToString:@" " intoString:&ctcpCommand];
        if (![scanner isAtEnd]) {
            [scanner scanString:@" " intoString:NULL];
            [scanner scanUpToString:@"" intoString:&ctcpParameters];
        }

        [array addObject:[[event arguments] objectAtIndex:0]];
        [array addObject:ctcpCommand];
        if (ctcpParameters != nil)
            [array addObject:ctcpParameters];
        
        ctcpEvent = [[IRCEvent alloc] initWithPrefix:[event prefix]
                                             command:@"CTCP"
                                           arguments:[[array copy] autorelease]];

        [_eventProcessor dispatchEvent:[ctcpEvent autorelease]];
    }
}

- (void)serverDidCTCP:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:1] isEqualToString:@"VERSION"]) {
        NSString *version, *osName = nil;

        switch ([[NSProcessInfo processInfo] operatingSystem]) {
            case NSWindowsNTOperatingSystem:	osName = @"Windows NT";		break;
            case NSWindows95OperatingSystem:	osName = @"Windows 95";		break;
            case NSSolarisOperatingSystem:	osName = @"Solaris";		break;
            case NSHPUXOperatingSystem:		osName = @"HP/UX";		break;
            case NSMACHOperatingSystem:		osName = @"Mach";		break;
            case NSSunOSOperatingSystem:	osName = @"Windows NT";		break;
            case NSOSF1OperatingSystem:		osName = @"OSF/1";		break;
            default:				osName = @"Unknown";		break;
        }

        version = [NSString stringWithFormat:@"%@ for OPENSTEP on %@ - v%@ (IOF %@)",
            [[NSProcessInfo processInfo] processName],
            osName,
            [IRCApp applicationVersion],
            IRCFrameworkVersion];

        [self sendEvent:[IRCEvent ctcpReplyEventWithTarget:[[event prefix] nicknameComponent]
                                                   command:@"VERSION"
                                                   message:version]];
    }
}

- (void)serverDidPING:(IRCServer *)server event:(IRCEvent *)event
{
    [self sendEvent:[IRCEvent pongEventWithServer:[[event arguments] objectAtIndex:0]]];
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] host: \"%@\" channels: %@ users: %@>",
        [[self class] description], self,
        _serverHost, [_channelDictionary description], [_users description]];
}

@end
