/* IRCUser.m created by dwy on Tue 02-Jun-1998 */

#import "IRC.h"

NSString *IRCUserVoicePrefix = @"+";
NSString *IRCUserChannelOperatorPrefix = @"@";
NSString *IRCUserServerOperatorPrefix = @"*";

DEFINE_NSSTRING(IRCUserDidChangeNotification);
DEFINE_NSSTRING(IRCUserDidChangeNicknameNotification);

#define IRCUserClassVersion		1

@implementation IRCUser

+ (void)initialize
{
    if (self == [IRCUser class])
        [self setVersion:IRCUserClassVersion];
}

+ (IRCUser *)userWithNickname:(NSString *)name server:(IRCServer *)server
{
    if (![[server users] containsUserNamed:[name stringByRemovingUserModeCharacters]])
        [server addUser:[[[self alloc] initWithNickname:name server:server] autorelease]];

    return [[server users] userNamed:[name stringByRemovingUserModeCharacters]];
}

- (id)initWithNickname:(NSString *)name server:(IRCServer *)server
{
    if (self = [super init]) {
        _server = server;
        _nickname = name;

        if (name == nil || server == nil)
            [NSException raise:NSInvalidArgumentException
                        format:@"nil name(%@) or server(%d) for initWithNickname:",
                name, [server description]];

        // strip channel/server status characters
        while ([_nickname length] > 0) {
            if ([_nickname hasPrefix:IRCUserVoicePrefix] ||
                [_nickname hasPrefix:IRCUserChannelOperatorPrefix])
                _nickname = [_nickname substringFromIndex:1];
            else if ([_nickname hasPrefix:IRCUserServerOperatorPrefix]) {
                _isServerOperator = YES;
                _nickname = [_nickname substringFromIndex:1];
            }
            else
                break;
        }

        [_nickname retain];
        [[_server eventProcessor] addResponder:self];
        [_server sendEvent:[IRCEvent whoisEventWithName:_nickname]];
    }

    return self;
}

- (void)dealloc
{
    [_nickname release];
    [_userName release];
    [_hostName release];
    [_userDescription release];
    [_serverName release];
    [_serverInfo release];
    [_channelNames release];
    [_awayMessage release];
    [_modes release];
    [_signOnDate release];

    [super dealloc];
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] name: %@ (%@@%@) info: \"%@\" serverName: %@ serverInfo: \"%@\" channelNames: %@ awayMessage: \"%@\" modes: %@>",
        [[self class] description], self,
        _nickname, _userName, _hostName, _userDescription,
        _serverName, _serverInfo,
        [_channelNames description],
        _awayMessage, _modes];
}

- (NSComparisonResult)compare:(IRCUser *)otherUser
{
    return [_nickname isEqualToStringIgnoringCase:[otherUser nickname]];
}

- (NSString *)nickname
{
    return [[_nickname copy] autorelease];
}

// JKDocument compatibility
- (NSString *)name
{
    return [self nickname];
}

- (IRCServer *)server
{
    return _server;
}

- (NSString *)userName
{
    return [[_userName copy] autorelease];
}

- (NSString *)hostName
{
    return [[_hostName copy] autorelease];
}

- (NSString *)userDescription;
{
    return [[_userDescription copy] autorelease];
}

- (void)setUserDescription:(NSString *)userDescription
{
    [_userDescription release];
    _userDescription = [userDescription retain];
    [self registerInternalChangeWithNotificationCenter];
}

- (void)setNickname:(NSString *)nickname
{
    [_nickname release];
    _nickname = [nickname retain];
    [self registerInternalChangeWithNotificationCenter];
}

- (NSString *)serverName
{
    return [[_serverName copy] autorelease];
}

- (NSString *)serverInfo
{
    return [[_serverInfo copy] autorelease];
}

- (NSArray *)channelNames
{
    return [[_channelNames copy] autorelease];
}

- (BOOL)isAway
{
    return (_awayMessage != nil);
}

- (NSString *)awayMessage
{
    return [[_awayMessage copy] autorelease];
}

- (NSString *)modes
{
    return [[_modes copy] autorelease];
}

- (NSCalendarDate *)idleSinceDate
{
    return [[_idleSinceDate copy] autorelease];
}

- (NSCalendarDate *)signOnDate
{
    return [[_signOnDate copy] autorelease];
}

- (BOOL)isServerOperator
{
    return _isServerOperator;
}

- (id)copyWithZone:(NSZone *)zone
{
    return [self retain];
}

// away message
- (void)serverDid301:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:1] isEqualToStringIgnoringCase:_nickname]) {
        _awayMessage = [[[event arguments] objectAtIndex:2] retain];
        
        [self registerInternalChangeWithNotificationCenter];
    }
}

// user@host, description
- (void)serverDid311:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:1] isEqualToStringIgnoringCase:_nickname]) {
        _userName = [[[event arguments] objectAtIndex:2] retain];
        _hostName = [[[event arguments] objectAtIndex:3] retain];
        _userDescription = [[[event arguments] objectAtIndex:5] retain];
        
        [self registerInternalChangeWithNotificationCenter];
    }
}

// server name and info
- (void)serverDid312:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:1] isEqualToStringIgnoringCase:_nickname]) {
        _serverName = [[[event arguments] objectAtIndex:2] retain];
        _serverInfo = [[[event arguments] objectAtIndex:3] retain];

        [self registerInternalChangeWithNotificationCenter];
    }
}

// idle time
- (void)serverDid317:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:1] isEqualToStringIgnoringCase:_nickname]) {
        _idleSinceDate = [[NSDate dateWithTimeIntervalSinceNow:(0-[[[event arguments] objectAtIndex:2] intValue])] retain];
        
        [self registerInternalChangeWithNotificationCenter];
    }
}

// on channels
- (void)serverDid319:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event arguments] objectAtIndex:1] isEqualToStringIgnoringCase:_nickname]) {
        NSEnumerator *nameEnumerator = [[[[event arguments] objectAtIndex:2] componentsSeparatedByString:@" "] objectEnumerator];
        NSString *name;

        [_channelNames release];
        _channelNames = [[NSMutableArray alloc] init];
        while (name = [nameEnumerator nextObject])
            if ([name length] > 0)
                [_channelNames addObject:name];

        [self registerInternalChangeWithNotificationCenter];
    }
}

- (void)serverDidNICK:(IRCServer *)server event:(IRCEvent *)event
{
    if ([[[event prefix] nicknameComponent] isEqualToStringIgnoringCase:_nickname])
        [self setNickname:[[event arguments] objectAtIndex:0]];
}

- (void)serverWillTerminate:(NSNotification *)notification
{
    [[_server eventProcessor] removeResponder:self];
}

@end
