#import "JediKnight.h"

@implementation JKDocument

- (id)initWithServer:(IRCServer *)server target:(id)target;
{
    if (self = [super init]) {
        NSString *createNotificationName = [NSString stringWithFormat:@"%@DidCreateNotification", [[self class] description]];

        _server = [server retain];
        _target = [target retain];
         
        if (![NSBundle loadNibNamed:[[self class] description] owner:self])
            TCLog(self, _cmd, @"Can't load %@.nib: application may be damaged.", [[self class] description]);

        [window setFrameAutosaveName:@"JKDocument"];
        [window setFrameUsingName:@"JKDocument"];

        if ([NSApp mainWindow] != nil)
            [window setFrameTopLeftPoint:[window cascadeTopLeftFromPoint:[[NSApp mainWindow] frame].origin]];

        [[NSNotificationCenter defaultCenter] addObserver:self
                                                 selector:@selector(serverWillTerminate:)
                                                     name:IRCServerWillTerminateNotification
                                                   object:_server];

        [[_server eventProcessor] addResponder:self];

        _outputManager = [[JKOutputManager alloc] initWithTextView:textView];
        [[JKInputManager defaultInputManager] registerCommandName:@"me"
                                                           target:self
                                                         selector:@selector(textCommandME:)
                                                        isPrivate:YES];

        [[NSNotificationCenter defaultCenter] postNotificationName:createNotificationName
                                                            object:self];
    }

    return self;
}

- (void)dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];

    [_target release];
    
    [[_server eventProcessor] removeResponder:self];
    [_server release];
    [_outputManager release];

    [super dealloc];
}

- (void)serverWillTerminate:(NSNotification *)notification
{
    [[_server eventProcessor] removeResponder:self];
    [textField setEnabled:NO];
}

- (NSWindow *)window
{
    return window;
}

- (NSTextView *)textView
{
    return textView;
}

- (void)textViewUpdateFinished
{
    if (![window isKeyWindow])
        [window setDocumentEdited:YES];

    if ([window isMiniaturized])
        [window setMiniwindowImage:[NSImage imageNamed:@"channelDocumentIconChanged"]];
}

- (void)clearTextView:(id)sender
{
    [textView clear];
    [_outputManager setPrependNewline:NO];
}

- (void)textEntryFinished:(id)sender
{
    NSString *entry = [textField stringValue];
    NSEnumerator *lineEnumerator = nil;
    NSString *line = nil;

    if ([entry rangeOfString:@"\n"].location != NSNotFound)
        lineEnumerator = [[entry componentsSeparatedByString:@"\n"] objectEnumerator];
    else
        lineEnumerator = [[NSArray arrayWithObject:entry] objectEnumerator];

    while (line = [lineEnumerator nextObject]) {
        if ([line length] > 0) {
            switch ([[JKInputManager defaultInputManager] interpretCommand:line
                                                              privateOwner:self]) {
                case JKInterpretedNoCommand:
                    [_server sendEvent:[IRCEvent messageEventWithTarget:[_target name]
                                                                message:line]];
                    [_outputManager appendSpokenMessage:line];
                    break;

                case JKInterpretedInvalidCommand:
                    [_outputManager appendClientMessage:@"Invalid command: \"%@\"", line];
                    break;

                case JKInterpretedCommand:
                default:
                    break;
            }
        }
    }
    
    [textField setStringValue:@""];
    [window makeFirstResponder:textField];
}

- (void)textCommandME:(NSString *)action
{
    [_server sendEvent:[IRCEvent ctcpEventWithTarget:[_target name]
                                             command:@"ACTION"
                                             message:action]];
    [_outputManager appendActedAction:action actor:[[_server localUser] nickname]];
}

- (void)windowDidBecomeMain:(NSNotification *)notification
{
    NSString *mainNotificationName = [NSString stringWithFormat:@"%@DidBecomeMainNotification", [[self class] description]];

    [textView setSelectedRange:NSMakeRange(0, 0)];
    [window setDocumentEdited:NO];
    [textField selectText:nil];
    
    [[NSNotificationCenter defaultCenter] postNotificationName:mainNotificationName
                                                        object:self];
}

- (void)windowWillClose:(NSNotification *)notification
{
    NSString *mainNotificationName = [NSString stringWithFormat:@"%@DidBecomeMainNotification", [[self class] description]];
    NSString *closeNotificationName = [NSString stringWithFormat:@"%@DidCloseNotification", [[self class] description]];
    
    [[_server eventProcessor] removeResponder:self];

    if ([[notification object] isKeyWindow])
        [[NSNotificationCenter defaultCenter] postNotificationName:mainNotificationName
                                                            object:nil];

    [[NSNotificationCenter defaultCenter] postNotificationName:closeNotificationName
                                                        object:self];
}

- (void)windowWillMiniaturize:(NSNotification *)aNotification
{
    [window setMiniwindowTitle:[_target name]];
    [window setMiniwindowImage:[NSImage imageNamed:@"channelDocumentIcon"]];
}

- (void)orderFrontFindPanel:(id)sender
{
    [[JKFindPanel sharedInstance] orderFrontFindPanel:sender];
}

- (void)findNext:(id)sender
{
    [[JKFindPanel sharedInstance] findNext:sender];
}

- (void)findPrevious:(id)sender
{
    [[JKFindPanel sharedInstance] findPrevious:sender];
}

- (void)enterSelection:(id)sender
{
    NSRange range = [textView selectedRange];
    if (range.length) {
        [[JKFindPanel sharedInstance] setFindString:[[textView string] substringWithRange:range]];
    } else {
        NSBeep();
    }
}

- (void)jumpToSelection:(id)sender
{
    [textView scrollRangeToVisible:[textView selectedRange]];
}

@end
